#include "mainwindow.h"
#include <QApplication>
#include "dialog.h"
#include<QtCore>
#include <QDebug>
#include <QThread>
#include "c:\opencv\build\include\opencv2\highgui\highgui.hpp"
#include <string>
#include <cmath>
#include "c:\opencv\build\include\opencv\cv.h"
#include "c:\opencv\build\include\opencv2/nonfree/features2d.hpp"
#include "c:\opencv\build\include\opencv2/objdetect/objdetect.hpp"
#include "c:\opencv\build\include\opencv2/highgui/highgui.hpp"
#include "c:\opencv\build\include\opencv2/imgproc/imgproc.hpp"
#include <iostream>
#include "mythread.h"
#include "ui_mainwindow.h"
#include "ui_dialog.h"
#define DEBUG 1
#define PI 3.14159265
using namespace std;
using namespace cv;

static int init();
static void findH();
static void findDarts();
static void drawDarts();
static void findCenters();
static int findScore(Point P);
static void showImages();
static bool handleInput();
static void createScoreMask();
static bool preciseCenters();
static int getScores();
static void setup();
static int cam();
RNG rng(12345);
int delay = 0, key=0, i=0;
CvCapture *video = NULL;
bool maskOn = true;
Mat board;
Mat originalBoard;
Mat mask;
Mat refMask;
Mat frame;
Mat grey;
Mat HSVimg;
Mat colorFiltered;
Mat H;
Mat drawing;
int slices[] = { 6, 10, 15, 2, 17, 3, 19, 7, 16, 8, 11, 14, 9, 12, 5, 20, 1, 18, 4, 13};
int multipliers[] = {1, 2, 3, 4};
Point bullseye(380,227);
double innerRadius = 23;
double outerRadius = 128;
vector<vector<Point>> darts;
vector<vector<Point>> dartCenters;
vector<vector<Point> > contours;
vector<Point> averageCenters;
vector<Vec4i> hierarchy;


MainWindow* pw;
QString Aname;
QString Bname;
int FScore;
int main(int argc, char *argv[])
{
    QApplication a(argc,argv);

    MainWindow w;
    pw  = &w;
    w.toRun = &cam;
    w.setAttribute(Qt::WA_QuitOnClose);

      Dialog* DBox;
      DBox = new Dialog();
      DBox->setModal(true);
      DBox->exec();

     Aname = DBox->AName;
     Bname = DBox->BName;
     FScore = DBox->FScore;
     delete DBox;
     setup();
     w.show();
     a.setQuitOnLastWindowClosed(true);

    return exit(a.exec());
}

static int cam()
{
    frame = cvQueryFrame(video);
    blur( frame, frame, Size(3,3) );
    //board = originalBoard.clone();
    while (!frame.empty())
    {
        darts.clear();
        findDarts();

        if(darts.size() == 1 )
        {
            drawDarts();
            findCenters();
            if(preciseCenters())
            {
                pw->CurrentScore =  getScores();
                dartCenters.clear();
                cvDestroyAllWindows();
                return 2;
            }
        }

        showImages();
        if(handleInput())
            break;
        /* load and check next frame*/
        frame = cvQueryFrame(video);
        if(frame.empty())
        {
            printf("error loading frame.\n");
            return 1;
        }
    }
}

static void findH()
{
    if(DEBUG)
        cout << "Finding Homography." << endl;
    SiftFeatureDetector detector(100);  // Detector for features

    vector<KeyPoint> srcFeatures;   // Detected key points on first image
    vector<KeyPoint> dstFeatures;
    detector.detect(mask,srcFeatures);
    detector.detect(refMask,dstFeatures);
    if(DEBUG)
        cout << "Features detected." << endl;
    // Extract descriptors of the features
    SiftDescriptorExtractor extractor;
    Mat srcDescriptors, dstDescriptors;
    extractor.compute(mask,  srcFeatures, srcDescriptors);
    extractor.compute(refMask , dstFeatures, dstDescriptors );
    if(DEBUG)
        cout << "Descriptors Extracted" << endl;
    // Match descriptors of 2 images (find pairs of corresponding points)
    FlannBasedMatcher matcher;       // Use FlannBasedMatcher matcher. It is better
    vector<DMatch> matches;
    matcher.match(srcDescriptors, dstDescriptors, matches);

    // Extract pairs of points
    vector<int> pairOfsrcKP(matches.size()), pairOfdstKP(matches.size());
    for( size_t i = 0; i < matches.size(); i++ ){
        pairOfsrcKP[i] = matches[i].queryIdx;
        pairOfdstKP[i] = matches[i].trainIdx;
    }
    if(DEBUG)
        cout << "Matches Found." << endl;
    vector<Point2f> sPoints; KeyPoint::convert(srcFeatures, sPoints,pairOfsrcKP);
    vector<Point2f> dPoints; KeyPoint::convert(dstFeatures, dPoints,pairOfdstKP);

    // Matched pairs of 2D points. Those pairs will be used to calculate homography
    Mat src2Dfeatures;
    Mat dst2Dfeatures;
    Mat(sPoints).copyTo(src2Dfeatures);
    Mat(dPoints).copyTo(dst2Dfeatures);
    if(DEBUG)
        cout << "Calculating Homography..." << endl;
    // Calculate homography
    vector<uchar> outlierMask;
    H = findHomography( src2Dfeatures, dst2Dfeatures, outlierMask, RANSAC,3);
    if(DEBUG)
        cout << "Homography Found." << endl;

}

static int init()
{
    video =	cvCaptureFromCAM(1);
    if(!video)
    {
        cout << "Could not open camera" << endl;
        cvWaitKey(0);
        return -1;
    }
    if(DEBUG)
        cout << "Video Capture started." << endl;
    refMask = imread("maskStraight.jpg");
    if(!refMask.data)
    {

        cout << "Could not read maskStraight.jpg" << endl;
        cvWaitKey(0);
        return -2;
    }
    if(DEBUG)
        cout << "Reference mask loaded." << endl;
    mask = imread("mask.jpg");
    if(!mask.data)
    {
        cout << "Could not read mask.jpg" << endl;
        cvWaitKey(0);
        return -3;
    }
    if(DEBUG)
        cout << "Alignment mask loaded." << endl;

    return 1;
}

static void findDarts()
{
    if(DEBUG)
        cout << "Finding Darts... " << endl;
    cvtColor(frame,HSVimg, CV_BGR2HSV);
    inRange(HSVimg, Scalar(35,92,58), Scalar(65,256,256),colorFiltered);
    findContours( colorFiltered, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE, Point(0, 0) );
    drawing =  Mat::zeros( colorFiltered.size(), CV_8UC3 );

    int contArea;
    for( unsigned int i = 0; i< contours.size(); i++ )
    {
        contArea = (int)contourArea(contours[i]);

        if((contArea > 600) && (contArea < 2000))
            darts.push_back(contours[i]);
    }

    if(DEBUG)
        cout << darts.size() << " darts found." << endl;
    dartCenters.resize(darts.size());
}

static void drawDarts()
{
    Scalar color = Scalar(0,0,0);
    for(unsigned int i = 0; i < darts.size(); i++)
    {
        drawContours( frame, darts, i, color, 2, 8, hierarchy, 0, Point() );
        drawContours( drawing, darts, i, Scalar(0,255,0), 2,8, hierarchy, 0, Point() );
    }
}

static void findCenters()
{
    if(DEBUG)
        cout << "Finding Centers..." << endl;

    Mat grayDarts;
    cvtColor(drawing,grayDarts,CV_BGR2GRAY);
    vector<Vec3f> circles;
    /// Apply the Hough Transform to find the circles
    HoughCircles( grayDarts, circles, CV_HOUGH_GRADIENT, 1, grayDarts.rows/8, 200, 17, 0, 0 );

    for(unsigned int i = 0; i < darts.size(); i++)
    {
        Point leftmost(800,0);
        for(unsigned int j = 0; j < darts[i].size(); j++)
            if(darts[i][j].x < leftmost.x)
                leftmost = darts[i][j];


        int minDist = 1000;
        int radius;
        Point center(0,0);
        for( size_t k = 0; k < circles.size(); k++ )
        {
            Point tempCenter(cvRound(circles[k][0]), cvRound(circles[k][1]));
            int dist = (tempCenter.x - leftmost.x) * (tempCenter.x - leftmost.x) + (tempCenter.y - leftmost.y) * (tempCenter.y - leftmost.y);
            //circle( drawing, tempCenter, 3, Scalar(255,0,0), -1);
            // circle outline
            //circle( drawing, tempCenter, cvRound(circles[k][2]), Scalar(255,0,0), 3);
            if(dist > 0 && dist < minDist && dist < 20*20)
            {
                radius = cvRound(circles[k][2]);
                minDist = dist;
                center = tempCenter;
            }
        }
        if(center.x != 0 && center.y != 0)
            dartCenters.at(i).push_back(Point(center.x,center.y));

    //	cout << "Dart " << i << " Dist " << minDist << endl;
    //	cout << "Leftmost " << leftmost.x << " Center " << center.x << "," << center.y << endl;
    /// Draw the circles detected

    // circle center
    //circle( drawing, center, 3, Scalar(0,0,255), -1);
    // circle outline
    //circle( drawing, center, radius, Scalar(0,0,255), 3);
    circle(frame, center,3,Scalar(0,0,0), -1);
    //circle( drawing, leftmost, 1, Scalar(0,0,255), -1);
    /*
    Mat centroidMat = Mat::zeros( board.size(), CV_8UC3 );
    circle(centroidMat, center,3,Scalar(0,255,255),-1);
    warpPerspective(centroidMat,centroidMat,H,refMask.size(),INTER_LINEAR,BORDER_CONSTANT);
    board = board + centroidMat;*/
    }


}

static void showImages()
{
  /// Show in a window
//	if(drawing.data)
//		imshow( "Contours", drawing );
   maskOn = pw->getblnMask();
    if(maskOn) frame = frame + mask;

    if(frame.data)
    {
       // imshow("Original", frame);
        pw->image = frame;
    }
//	if(board.data)
    //	imshow("Projected Centroid", board);
}

static bool handleInput()
{
  /* wait delay and check for the quit key */
  key = cvWaitKey(10);
  if(key=='q') return true;
  if(key=='s')
    cvWaitKey(0);
  //else if(key == 'm')maskOn = !maskOn;
  return false;
}

static void createScoreMask()
{
    cvtColor(refMask,refMask,CV_BGR2GRAY);
    threshold(refMask,refMask,50,255,0);
}

static int findScore(Point p)
{
    double angle;
    angle = atan2((float)(bullseye.y - p.y),(float)(bullseye.x - p.x )) * 180/PI + 181;
    int index =(int)ceil( (angle-9)/18 );
    if(index > 19)
        index = 0;

    int baseScore = slices[index];
    int mult;
    double bullDistance = sqrt((double)(p.x - bullseye.x) * (p.x - bullseye.x) + (p.y - bullseye.y - 5) * (p.y - bullseye.y - 5));
    double centerDistance = sqrt((double)(p.x - bullseye.x) * (p.x - bullseye.x) + (p.y - bullseye.y) * (p.y - bullseye.y));

    if(bullDistance < sqrt(innerRadius * innerRadius))
        return 100;
    else if(refMask.at<char>(p) != 0)
    {
        if(centerDistance < sqrt(outerRadius * outerRadius))
            mult = multipliers[2];
        else
            mult = multipliers[1];
    }
    else
        mult = multipliers[0];
    return baseScore * mult;
}

static bool preciseCenters()
{
    if(DEBUG)
        cout << "Smoothing Centers..." << endl;
    for(unsigned int i = 0; i < dartCenters.size(); i++)
        if(dartCenters[i].size() < 10)
        {
            if(DEBUG)
                cout << "Not enough centers." << endl;
            return false;
        }

    for(unsigned int i = 0; i < dartCenters.size(); i++)
    {
        int x = 0, y = 0;
        for(unsigned int j = 0; j < dartCenters[i].size(); j++)
        {
            x+= dartCenters[i][j].x;
            y+= dartCenters[i][j].y;
        }

        x /= dartCenters[i].size();
        y /= dartCenters[i].size();
        dartCenters[i].push_back(Point(x,y));
    }

    cout << "Centers founds." << endl;
    return true;


}

int  getScores()
{
    if(DEBUG)
        cout << "Getting Scores..." << endl;
    vector<Point2f> rawCenters;

    for(unsigned int i = 0; i < dartCenters.size(); i++)
        rawCenters.push_back(dartCenters[i][dartCenters[i].size()-1]);
    cout << "Size: " << rawCenters.size();
    cout << rawCenters << endl;
    vector<Point2f> warpedCenters;
    warpedCenters.resize(rawCenters.size());
    if(DEBUG)
        cout << "Transforming centers... " << endl;
    perspectiveTransform(rawCenters,warpedCenters,H);

    if(DEBUG)
        cout << "Centers Transformed." << endl;
    for(unsigned int i = 0; i < warpedCenters.size(); i++)
        cout << "Dart " << i << " Score: " << findScore(warpedCenters[i]) << endl;

    return findScore(warpedCenters[0]);
}

static void setup(void){
    pw->SetNames(Aname,Bname);
    pw->setScore(FScore);
    int status;
    status = init();
    findH();
    createScoreMask();

}

